/*
  'MKUltra' kwin client

  Copyright (C) 2004 Gabor Bernath <gabor@scanguru.hu>
  Port to KDE 3.2.x 
    Copyright (C) 2002 Shamyl Zakariya <zakariya@earthlink.net>
	Based on excellent Web style, 
	  Copyright (C) 2001 Rik Hemsley (rikkus) <rik@kde.org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
*/

#include <qcursor.h>
#include <qpainter.h>
#include <qtooltip.h>
#include <qapplication.h>

#include "MKUltraButton.h"
#include "MKUltra.h"

namespace MKUltra {

/**
	Returns a new color @amount/255 from a -> b
*/
QColor g_tweenColor(const QColor &a, const QColor &b, int amount)
{
	int dR = b.red() - a.red();
	int dG = b.green() - a.green();
	int dB = b.blue() - a.blue();

	return QColor( 	a.red() + ((dR * amount) >> 8), 
					a.green() + ((dG * amount) >> 8), 
					a.blue() + ((dB * amount) >> 8));
}


MKUltraButton::MKUltraButton(QWidget * parent, const QString& tip, MKUltraClient* deco)
  : QButton (parent, 0, 0),
    mouseOver_  (false),
    mouseDown_  (false),
    position_   (Mid),
    shape_      (false),
    deco_       (deco)
{
  setTipText(tip);
  setCursor(ArrowCursor);
  setSizePolicy(QSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum));
  setBackgroundMode(NoBackground);
}

MKUltraButton::~MKUltraButton()
{
  // Empty.
}

  void
MKUltraButton::setShape(bool b)
{
  shape_ = b;
  repaint();
}

  void
MKUltraButton::mousePressEvent(QMouseEvent * e)
{
  mouseDown_ = true;
  repaint();
  QButton::mousePressEvent(e);
}

  void
MKUltraButton::mouseReleaseEvent(QMouseEvent * e)
{
  mouseDown_ = false;
  repaint();

  KDecorationFactory* f = deco_->factory();
  if (rect().contains(e->pos()))
  {
    clickEvent(e->button());
  }
  if( !f->exists( deco_ )) // decoration was destroyed
    return;
  QButton::mouseReleaseEvent(e);
}

  void
MKUltraButton::enterEvent(QEvent * e)
{
  mouseOver_ = true;
  repaint();
  QButton::enterEvent(e);
}

  void
MKUltraButton::leaveEvent(QEvent * e)
{
  mouseOver_ = false;
  repaint();
  QButton::leaveEvent(e);
}

  void
MKUltraButton::paintEvent(QPaintEvent *)
{
  QPen highlightPen;
  if (mouseDown_)
    highlightPen = QPen(colorGroup().light());

  else
  {
    if (mouseOver_)
      highlightPen = QPen(colorGroup().highlight());
    else
      highlightPen = QPen(NoPen);
  }

  QPainter p(this);

  p.fillRect(rect(), KDecoration::options()->color(KDecorationOptions::ColorTitleBar, deco_->isActive())); 

  //tint colors for titlebar
  QColor bodyLightColor = colorGroup().background().light();
  QColor topRectLightColor = KDecoration::options()->color(KDecorationOptions::ColorTitleBar, deco_->isActive()).light(120);
  QColor topRectDarkColor = KDecoration::options()->color(KDecorationOptions::ColorTitleBar, deco_->isActive()).dark(120);

  switch ( position_ )
  {
    case Left:
      {
			// Draw edges.
			p.setPen(topRectLightColor);
			p.drawLine(1, 1, width(), 1);
			p.drawLine(1, 1, 1, height());

			p.setPen(bodyLightColor);
			p.drawLine(1, height() - 1, width(), height() - 1);

			p.setPen(topRectDarkColor);
			p.drawLine(1, height() - 2, width(), height() - 2);

			p.setPen(Qt::black);

			p.drawLine(0, 0, width(), 0);
			p.drawLine(0, 1, 0, height() - 1);


			if (shape_)
			{
				p.drawPoint(3, 1);
				p.drawPoint(4, 1);
				p.drawPoint(2, 2);
				p.drawPoint(1, 3);
				p.drawPoint(1, 4);

				p.setPen(topRectLightColor);
				p.drawPoint(3, 2);
				p.drawPoint(4, 2);
				p.drawPoint(2, 3);
				p.drawPoint(2, 4);
				p.drawPoint(2, 4);
			}

      }
      
	  break;

    case Right:
      {
			// Draw edges.
			p.setPen(topRectLightColor);
			p.drawLine(0, 1, width(), 1);

			p.setPen(bodyLightColor);
			p.drawLine(0, height() - 1, width() - 3, height() - 1);

			p.setPen(topRectDarkColor);
			p.drawLine(width() - 2, 1, width() - 2, height() - 2);
			p.drawLine(0, height() - 2, width() - 1, height() - 2);

			p.setPen(Qt::black);
			p.drawLine(0, 0, width(), 0);
			p.drawLine(width() - 1, 1, width() - 1, height() - 1);

			if (shape_)
			{
				p.drawPoint(width() - 5, 1);
				p.drawPoint(width() - 4, 1);
				p.drawPoint(width() - 3, 2);
				p.drawPoint(width() - 2, 3);
				p.drawPoint(width() - 2, 4);

				p.setPen(topRectLightColor);
				p.drawPoint(width() - 5, 2);
				p.drawPoint(width() - 4, 2);

				p.setPen(topRectDarkColor);
				p.drawPoint(width() - 3, 3);
				p.drawPoint(width() - 3, 4);

			}
	  
      }
      
	  break;

    case Mid:
    default:
      {
			// Draw edge.
			p.setPen(topRectLightColor);
			p.drawLine(0, 1, width(), 1);

			p.setPen(bodyLightColor);
			p.drawLine(0, height() - 1, width(), height() - 1);

			p.setPen(topRectDarkColor);
			p.drawLine(0, height() - 2, width(), height() - 2);


			p.setPen(Qt::black);
			p.drawLine(0, 0, width(), 0);
      }

      break;
  }

  // Draw icon.

  QPoint center(rect().center());

  int bwby2(bitmap_.width() / 2);    // Bitmap Width BY 2
  int bhby2(bitmap_.height() / 2);   // Bitmap Height BY 2

  p.setBrush(NoBrush);

  QColor titleBarFontColor = KDecoration::options()->color(KDecorationOptions::ColorFont, deco_->isActive());
  if (mouseOver_)
  	  p.setPen(titleBarFontColor);
  else
  	  p.setPen( g_tweenColor( KDecoration::options()->color(KDecorationOptions::ColorTitleBar, deco_->isActive()), titleBarFontColor, 96 ) );
  
  p.drawPixmap(center.x() - bwby2 + 1, center.y() - bhby2 + 1, bitmap_);

}

  QSize
MKUltraButton::sizeHint() const
{
  return QSize(16, 16);
}

  QSize
MKUltraButton::minimumSizeHint() const
{
  return QSize(14, 14);
}

  void
MKUltraButton::setBitmap(const QBitmap & b)
{
  bitmap_ = b;
  repaint();
}

  void
MKUltraButton::setPosition(Position p)
{
  if (  QApplication::reverseLayout() )
  {
      if ( p == Left )
          position_ = Right;
      else if ( p == Right )
          position_ = Left;
  }
  else
      position_ = p;
  repaint();
}

  void
MKUltraButton::resizeEvent(QResizeEvent *)
{
  repaint();
}

void
MKUltraButton::setTipText(const QString &tip) {
  if(KDecoration::options()->showTooltips()) {
    QToolTip::remove(this );
    QToolTip::add(this, tip );
  }
}

}

#include "MKUltraButton.moc"
// vim:ts=2:sw=2:tw=78:set et:
